@@ -1,11 +1,13 @@
 """A collection of functions that assist in validation/comparison of data and conditions.
 """
-from typing import List, Union
+from typing import List, Union, Callable, Type, Iterable
 
 import numpy as np
 import pandas as pd
+from pandas.core.dtypes.common import is_categorical
 
 from helpsk.exceptions import *  # pylint: disable=wildcard-import,unused-wildcard-import
+from helpsk.utility import suppress_warnings
 
 
 def any_none_nan(values: Union[List, np.ndarray, pd.Series, pd.DataFrame]) -> bool:
@@ -154,8 +156,10 @@ def assert_true(condition: bool, message: str = 'Condition Not True') -> None:
     """Raises an HelpskAssertionError if `condition` is not True
 
     Args:
-        condition: bool
-            Something that evalualates to True/False
+        condition:
+            Something that evaluates to True/False
+        message:
+            Message passed to the HelpskAssertionError
     """
     if not isinstance(condition, (bool, np.bool_)):
         raise HelpskParamTypeError('condition should be boolean')
@@ -169,7 +173,9 @@ def assert_false(condition: bool, message: str = 'Condition True') -> None:
 
     Args:
         condition: bool
-            Something that evalualates to True/False
+            Something that evaluates to True/False
+        message:
+            Message passed to the HelpskAssertionError
     """
     if not isinstance(condition, (bool, np.bool_)):
         raise HelpskParamTypeError('condition should be boolean')
@@ -178,6 +184,56 @@ def assert_false(condition: bool, message: str = 'Condition True') -> None:
         raise HelpskAssertionError(message)
 
 
+def iterables_are_equal(iterable_a: Iterable, iterable_b: Iterable) -> bool:
+    """Compares the equality of the values of two iterables.
+
+    This function will generally give the same result as list equality (e.g. `[x, y, z] == [x, y, z]`)
+    However, in some strange scenarios, `==` will return `False` where it doesn't seem like it should
+
+    For example:
+
+    ```
+    temp = pd.DataFrame({'col_a': [np.nan, 1.0]})
+    temp.col_a.tolist() == [np.nan, 1.0]  # returns False. Why??
+    iterables_are_equal(temp.col_a.tolist(), [np.nan, 1])  # returns True
+    [np.nan, 1.0] == [np.nan, 1.0]  # returns True
+
+
+    Also, when comparing a series with an ordered Categorical when the values are the same,
+    pd.Series.equals() will return False if the categories have different order. But we only care if the
+    values are the same, so this function will return True.
+    ```
+
+    Args:
+        iterable_a:
+            an iterable to equate to iterable_b
+        iterable_b:
+            an iterable to equate to iterable_a
+
+    Returns:
+        True if iterable_a is equal to iterable_b
+    """
+    # seems to be confusion and inconsistencies across stack overflow on how to properly check for category
+    # so this might be overkill but not exactly sure
+    # def is_categorical(series):
+    #     if isinstance(series, (pd.Categorical, pd.CategoricalDtype)):
+    #         return True
+    #     if isinstance(series, pd.Series):
+    #         return series.dtype.name == 'category'
+    #     return False
+
+    with suppress_warnings():
+        # if either list-like structure is categorical, then we need to convert both to unordered categorical
+        if is_categorical(iterable_a) or is_categorical(iterable_b):
+            iterable_a = pd.Categorical(iterable_a, ordered=False)
+            iterable_b = pd.Categorical(iterable_b, ordered=False)
+        else:
+            iterable_a = pd.Series(iterable_a)
+            iterable_b = pd.Series(iterable_b)
+
+        return iterable_a.equals(iterable_b)
+
+
 def dataframes_match(dataframes: List[pd.DataFrame],
                      float_tolerance: int = 6,
                      ignore_indexes: bool = True,
@@ -241,7 +297,7 @@ def assert_dataframes_match(dataframes: List[pd.DataFrame],
                             ignore_column_names: bool = True,
                             message: str = 'Dataframes do not match') -> None:
     """
-    Raises an assertion error
+    Raises an assertion error if dataframes don't match.
 
     Args:
         dataframes:
@@ -265,3 +321,55 @@ def assert_dataframes_match(dataframes: List[pd.DataFrame],
                             ignore_indexes=ignore_indexes,
                             ignore_column_names=ignore_column_names):
         raise HelpskAssertionError(message)
+
+
+def is_close(value_a: float, value_b: float, tolerance: float = 0.000001) -> bool:
+    """Tests whether or not value_a and value_b are "close" (i.e. within the `tolerance` after subtracting)
+
+    Args:
+        value_a:
+            numeric value to test
+        value_b:
+            numeric value to test
+        tolerance:
+            the maximum difference (absolute value) allowed between value_a and value_b
+    Returns:
+          True if values are within specified tolerance
+    """
+    return abs(value_a - value_b) <= tolerance
+
+
+def assert_is_close(value_a: float, value_b: float, tolerance: float = 0.000001):
+    """Raises an assert error if value_a and value_b are not "close" (see documentation of `is_close()`
+    function).
+
+    Args:
+          value_a:
+              numeric value to test
+          value_b:
+              numeric value to test
+          tolerance:
+              number of digits to round to
+    """
+    if not is_close(value_a=value_a, value_b=value_b, tolerance=tolerance):
+        raise HelpskAssertionError(f"`{value_a}` and `{value_b}` are not within a tolerance of `{tolerance}`")
+
+
+def raises_exception(function: Callable, exception_type: Type = None) -> bool:
+    """Returns True if `function` raises an Exception; returns False if `function` runs without raising an
+    Exception.
+    Args:
+        function:
+            the function which does or does not raise an exception.
+        exception_type:
+            if `exception_type` is provided, `raises_exception` returns true only if the `function` argument
+            raises an Exception **and** the exception has type of `exception_type`.
+    """
+    try:
+        function()
+        return False
+    except Exception as exception:  # pylint: disable=broad-except
+        if exception_type:
+            return isinstance(exception, exception_type)
+
+        return True
